; oversampling in Csound to avoid aliasing in audio processing
; Oeyvind Brandtsegg 2020 obrandts@gmail.com

<CsoundSynthesizer>
<CsOptions>
</CsOptions>
<CsInstruments>

sr = 48000
ksmps = 32
nchnls = 2
0dbfs = 1

; the audio process to be oversampled
opcode Oversampled_proc, a, aiP
 ; audio in, oversampling factor, optional k-rate parameter for the audio process
 a1, ioversampl, kparm xin
 a1 = tanh(a1*kparm)
 xout a1
endop

; the audio process to be oversampled
opcode Oversampled_proc, a, aiP
 ; audio in, oversampling factor, optional k-rate parameter for the audio process
 a1, ioversampl, kparm xin
 kfq = kparm/ioversampl ; if the upsampled audio process requires a frequency parameter,
   ; this value needs to be scaled relative to the new sampling rate
 ; single sideband modulation
 aSin, aCos	hilbert	a1
 aModSin oscili	1, kfq, -1, 0.0
 aModCos oscili	1, kfq, -1, 0.25
 aMod1 = aSin * aModCos
 aMod2 = aCos * aModSin
 aSum	= aMod1 + aMod2
 aDiff = aMod1 - aMod2
 xout aDiff
endop

; oversampling of audio, to allow processing above Nyquist
; interpolation on upsampling, convolution as lowpass before decimation
; Oeyvind Brandtsegg 2020 obrandts@gmail.com
opcode Oversample, a, aiP
 ; audio in, oversampling factor, optional k-rate parameter for the audio process
 a1, ioversampl, kparm xin
 kaudiobuf[] init ksmps
 kaudiobuf shiftin a1
 kaudioup[] init ksmps*ioversampl

 ; upsample by zero stuffing and interpolation
 kindex = 0
 kprev init 0
 while kindex < ksmps do
  kaudioup[kindex*ioversampl] = kprev
  kindex2 = 1
  while kindex2 < ioversampl do
    kaudioup[kindex*ioversampl+kindex2] = kprev*(1-(kindex2/ioversampl))+kaudiobuf[kindex]*(kindex2/ioversampl)
    kindex2 += 1
  od
  kprev = kaudiobuf[kindex]
	kindex += 1
 od

; set lowpass filter coefficients
 #include "filter_coefs.inc"
 ifilter = ioversampl == 2 ? iUp2 : iUp4
 ifilter = ioversampl == 8 ? iUp8 : ifilter
 print ftlen(ifilter), ifilter

 ; oversampled audio processing loop
 ; with lowpass filter to minimize aliasing when decimation is done is the next step
 kindex = 0
 while kindex < ioversampl do
  aup shiftout kaudioup
  aup Oversampled_proc aup, ioversampl, kparm; insert your own processing here
  aup dconv aup, ftlen(ifilter), ifilter ; downsampling filter
  kaudioup shiftin aup ; write audio back to buffer
  kindex += 1
 od

 ; decimation
 kindex = 0
 while kindex < ksmps do
  kaudiobuf[kindex] = kaudioup[(kindex*ioversampl)]
  kindex += 1
 od
 a2 shiftout kaudiobuf
 xout a2
endop

instr 1
; test ooversampling
iamp = ampdbfs(-3)
a1 diskin2 "fox.wav", 1, 0, 1
;a1 poscil 1, 440

kparm line 0, p3, sr
ioversampl = 4
a2 Oversample a1, 8, kparm

; reference signal that has received the same processing, but not in upsampled form
aref Oversampled_proc a1, 1, kparm; insert your own processing here

outs aref*iamp, a2*iamp

endin

</CsInstruments>
<CsScore>
;	start	dur
i1	0	20
e
</CsScore>
</CsoundSynthesizer>
